import os
from unittest.mock import patch

from pytest_matchers.utils.warn import warn


@patch("pytest_matchers.utils.warn.warnings.warn")
def test_warn(mock_warn):
    with patch.dict(os.environ, {"PYTEST_MATCHERS_WARNINGS": "True"}):
        warn("A warning")
        mock_warn.assert_called_with("A warning", UserWarning)


@patch("pytest_matchers.utils.warn.warnings.warn")
def test_warn_with_category(mock_warn):
    with patch.dict(os.environ, {}, clear=True):
        warn("Another warning", DeprecationWarning)
        mock_warn.assert_called_with("Another warning", DeprecationWarning)


@patch("pytest_matchers.utils.warn.warnings.warn")
def test_warnings_disabled(mock_warn):
    with patch.dict(os.environ, {"PYTEST_MATCHERS_WARNINGS": "False"}):
        warn("A warning")
        mock_warn.assert_not_called()

    with patch.dict(os.environ, {"PYTEST_MATCHERS_WARNINGS": "True"}):
        warn("A warning")
        mock_warn.assert_called_with("A warning", UserWarning)



from pytest_matchers.matchers.different_value import DifferentValue


def test_create():
    matcher = DifferentValue()
    assert isinstance(matcher, DifferentValue)


def test_repr():
    matcher = DifferentValue()
    assert repr(matcher) == "To be a different value"
    matcher = DifferentValue()
    assert matcher == 3
    assert repr(matcher) == "To be different to 3"
    matcher = DifferentValue()
    assert matcher == "string"
    assert repr(matcher) == "To be different to 'string'"


def test_matches():
    matcher = DifferentValue()
    assert matcher == 3
    assert matcher != 3
    assert matcher == 4
    assert matcher == "string"
    matcher = DifferentValue()
    assert matcher == "string"
    assert matcher != "string"
    assert matcher == 3



from pytest_matchers import anything, is_number, is_string
from pytest_matchers.matchers import Dict


def test_create():
    matcher = Dict()
    assert isinstance(matcher, Dict)
    matcher = Dict({"key": 3})
    assert isinstance(matcher, Dict)
    matcher = Dict({"key": 3}, ["other"])
    assert isinstance(matcher, Dict)


def test_repr():
    matcher = Dict()
    assert repr(matcher) == "To be a dictionary"
    matcher = Dict({"key": 3})
    assert repr(matcher) == "To be a dictionary expecting 'key' to be 3"
    matcher = Dict({"key": 3, "other": "string"})
    assert (
        repr(matcher)
        == "To be a dictionary expecting 'key' to be 3 and expecting 'other' to be 'string'"
    )
    matcher = Dict(exclude=["key"])
    assert repr(matcher) == "To be a dictionary excluding 'key'"
    matcher = Dict({"key": 3}, ["other"])
    assert repr(matcher) == "To be a dictionary excluding 'other' and expecting 'key' to be 3"
    matcher = Dict({"key": is_string(), "any": anything(), "other": 3})
    assert (
        repr(matcher) == "To be a dictionary expecting 'key' to be string and "
        "expecting 'any' to be anything and expecting 'other' to be 3"
    )


def test_matches():
    matcher = Dict()
    assert matcher == {}  # pylint: disable=use-implicit-booleaness-not-comparison
    assert matcher == {"key": 3}
    assert matcher != [1, 2, 3]
    assert matcher != "string"
    matcher = Dict({"key": 3})
    assert matcher == {"key": 3}
    assert matcher == {"key": 3, "other": 4}
    assert matcher != {"key": 4}
    assert matcher != {"other": 3}
    matcher = Dict({"key": 3}, exclude=["other"])
    assert matcher == {"key": 3}
    assert matcher != {"key": 3, "other": 4}
    assert matcher == {"key": 3, "x": 5}
    matcher = Dict({"key": 3, "other": is_string()})
    assert matcher == {"key": 3, "other": "string"}
    assert matcher == {"key": 3, "other": "string", "x": [1, 2, 3]}
    assert matcher != {"key": 3, "other": 4}
    assert matcher != {"key": 3}
    matcher = Dict({1: "one", 8.9: [1, 2, 3], "cool": is_number()})
    assert matcher == {1: "one", 8.9: [1, 2, 3], "cool": 30.5}
    assert matcher != {1: "one", 8.9: [1, 2, 3], False: "false"}
    assert matcher == {1: "one", 8.9: [1, 2, 3], False: "false", "cool": 0}
    assert matcher != {1: "one", 8.9: [1, 2, 3], "cool": "string"}



from datetime import datetime
from random import random
from unittest.mock import MagicMock

from pytest_matchers import (
    anything,
    between,
    case,
    different_value,
    has_attribute,
    if_false,
    if_true,
    is_datetime,
    is_datetime_string,
    is_dict,
    is_instance,
    is_list,
    is_number,
    is_string,
    same_value,
)


def test_random_number():
    assert random() == is_number(min_value=0, max_value=1)
    assert random() == between(0, 1)


def test_dict_comparison():
    value = {
        "string": "string",
        "int": 3,
        "list": [1, 2, 3],
    }
    assert value == {
        "string": is_instance(str),
        "int": is_instance(int),
        "list": is_list(int, length=3),
    }
    assert value != {"string": is_instance(str), "int": is_instance(int)}


def test_mock_log_exception():
    def _logging_function(logging):
        try:
            raise ValueError("This is a test exception")
        except ValueError as error:
            logging.exception("Caught exception: %s", error)

    mock_logging = MagicMock()
    _logging_function(mock_logging)
    mock_logging.exception.assert_called_with(
        "Caught exception: %s",
        is_instance(ValueError) & has_attribute("args", ("This is a test exception",)),
    )


def test_datetime_comparison():
    def _datetime_function():
        return {
            "data": "some data",
            "created_at": datetime.now(),
            "created_at_string": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
        }

    expected_value = {
        "data": "some data",
        "created_at": is_datetime(min_value=datetime.now()),
        "created_at_string": is_datetime_string("%Y-%m-%d %H:%M:%S", min_value=datetime.now()),
    }
    first_data = _datetime_function()
    assert first_data == expected_value
    second_data = _datetime_function()
    assert second_data == expected_value
    assert first_data != second_data


def test_same_and_different_value_comparison():
    dynamic_value = None

    def _function(previous_value: float = 0):
        nonlocal dynamic_value
        if dynamic_value is None:
            dynamic_value = random()
        return {
            "static": "static",
            "dynamic_repeated": dynamic_value,
            "dynamic_different": previous_value + random(),
        }

    expected_value = {
        "static": "static",
        "dynamic_repeated": same_value() & is_number(min_value=0, max_value=1),
        "dynamic_different": different_value() & is_number(),
    }
    first_data = _function()
    assert first_data == expected_value
    second_data = _function(first_data["dynamic_different"])
    assert second_data == expected_value


def test_conditional_matchers():
    assert random() == if_true(
        lambda value: value > 0.5,
        is_number(min_value=0.5, inclusive=False),
        is_number(max_value=0.5),
    )
    assert random() == if_false(
        lambda value: value > 0.5,
        is_number(max_value=0.5),
        is_number(min_value=0.5, inclusive=False),
    )
    random_value = random()
    random_string = "low"
    if random_value > 0.5:  # pragma: no cover
        random_string = "medium"
    if random_value > 0.8:  # pragma: no cover
        random_string = "high"
    assert random_value == case(
        random_string,
        {
            "low": is_number(max_value=0.5),
            "medium": is_number(min_value=0.5, max_value=0.8, min_inclusive=False),
            "high": is_number(min_value=0.8, inclusive=False),
        },
    )
    matcher = if_true(is_string(), is_datetime_string("%Y-%m-%d"), is_datetime())
    assert "2021-01-01" == matcher
    assert datetime(2021, 1, 1) == matcher
    assert 3 != matcher
    assert random() != matcher
    assert "10/11/2029" != matcher


def test_compare_dictionaries():
    test_dict = {"a": 1, "b": [2, "string"], "c": True, "d": "true"}

    assert test_dict == {
        "a": 1,
        "b": is_list(length=2),
        "c": True,
        "d": anything(),
    }
    assert test_dict != {
        "a": 1,
        "b": is_list(length=2),
        "c": True,
    }
    assert test_dict == is_dict(
        {
            "a": 1,
            "b": is_list(length=2),
            "c": True,
        }
    )
    assert test_dict != is_dict(
        {
            "a": 1,
            "b": is_list(length=2),
            "c": True,
        },
        exclude=["d"],
    )
    assert test_dict == is_dict(
        {
            "a": 1,
            "b": is_list(length=2),
            "c": True,
        },
        exclude=["e"],
    )



from pytest_matchers.matchers.same_value import SameValue


def test_create():
    matcher = SameValue()
    assert isinstance(matcher, SameValue)


def test_repr():
    matcher = SameValue()
    assert repr(matcher) == "To be the same value"
    matcher = SameValue()
    assert matcher == 3
    assert repr(matcher) == "To be the same value as 3"
    matcher = SameValue()
    assert matcher == "string"
    assert repr(matcher) == "To be the same value as 'string'"


def test_matches():
    matcher = SameValue()
    assert matcher == 3
    assert matcher == 3
    assert matcher != 4
    assert matcher != "string"
    matcher = SameValue()
    assert matcher == "string"
    assert matcher == "string"
    assert matcher != 3



from pytest_matchers import is_number, is_string
from pytest_matchers.matchers import Case


def test_create():
    matcher = Case(3, {3: 3})
    assert isinstance(matcher, Case)
    matcher = Case(3, {3: is_number()})
    assert isinstance(matcher, Case)
    matcher = Case(4, {3: 3}, 3)
    assert isinstance(matcher, Case)
    matcher = Case(3, {3: 3}, 3)
    assert isinstance(matcher, Case)
    matcher = Case(3, {3: 3}, is_string())
    assert isinstance(matcher, Case)


def test_repr():
    matcher = Case(3, {3: 4})
    assert repr(matcher) == "To be 4 because case value is 3"
    matcher = Case(3, {3: is_number()})
    assert repr(matcher) == "To be a number because case value is 3"
    matcher = Case(4, {3: 4}, 5)
    assert repr(matcher) == "To be 5 because case value is 4"
    matcher = Case(4, {3: 3}, is_string())
    assert repr(matcher) == "To be string because case value is 4"
    matcher = Case(4, {3: 3})
    assert (
        repr(matcher) == "To never match because case value 4 is not expected "
        "and the default expectation is not set"
    )


def test_matches():
    matcher = Case(3, {3: 4, 4: "string"})
    assert matcher == 4
    assert matcher != 3
    assert matcher != 5
    assert matcher != "string"
    matcher = Case(3, {3: is_number()})
    assert matcher == 3
    assert matcher == 4
    matcher = Case(3, {3: is_string()}, 3)
    assert matcher == "3"
    assert matcher != 3
    matcher = Case(4, {3: is_number()}, is_string())
    assert matcher == "string"
    assert matcher != 3
    assert matcher != 4
    matcher = Case(4, {3: 3})
    assert matcher != 3
    assert matcher != 4
    assert matcher != "string"



from pytest_matchers import is_number, is_string
from pytest_matchers.matchers import If


def test_create():
    matcher = If(True)
    assert isinstance(matcher, If)
    matcher = If(False)
    assert isinstance(matcher, If)
    matcher = If(lambda x: x == 3)
    assert isinstance(matcher, If)
    matcher = If(lambda x: x == 3, 3)
    assert isinstance(matcher, If)
    matcher = If(lambda x: x == 3, 3, 4)
    assert isinstance(matcher, If)
    matcher = If(lambda x: x == 3, or_else=4)
    assert isinstance(matcher, If)


def test_repr():
    matcher = If(True)
    assert repr(matcher) == "To be anything"
    matcher = If(False, is_string(), is_number(int))
    assert repr(matcher) == "To be a number of 'int' instance"
    matcher = If(lambda x: x == 3)
    assert repr(matcher) == "To be anything if condition returns True else to be anything"
    matcher = If(lambda x: x == 3, 3)
    assert repr(matcher) == "To be 3 if condition returns True else to be anything"
    matcher = If(lambda x: x == 3, is_number(), 4)
    assert repr(matcher) == "To be a number if condition returns True else to be 4"
    matcher = If(lambda x: x == 3, or_else=is_string())
    assert repr(matcher) == "To be anything if condition returns True else to be string"


def test_match():
    assert If(True) == 3
    assert If(False) == 3
    matcher = If(True, is_number(), is_string())
    assert matcher == 3
    assert matcher == 4
    assert matcher != "string"
    matcher = If(False, is_number(), is_string())
    assert matcher != 3
    assert matcher != 4
    assert matcher == "string"
    matcher = If(lambda x: x > 3, is_number(float), is_number(int))
    assert matcher == 3
    assert matcher != 4
    matcher = If(lambda x: x + 2 == 5, 3)
    assert matcher == 3
    assert matcher == 4  # False can be anything
    assert matcher == "string"
    matcher = If(lambda x: x <= 3, 3, 4)
    assert matcher == 3
    assert matcher == 4
    assert matcher != "string"
    matcher = If(lambda x: x + 2 == 5, or_else=4)
    assert matcher == 3
    assert matcher == 4
    assert matcher != "string"
    matcher = If(5, is_number(), is_string())
    assert matcher == 5
    assert matcher != 4
    assert matcher == "string"
    matcher = If(is_string(), is_string(length=8), is_number())
    assert matcher == "string__"
    assert matcher != "string"
    assert matcher == 3
    assert matcher != [1]
